#
# Xsigo supplied xm command utilities
#
# Author: David Allison
#
# Copyright (c) 2007 Xsigo Systems Inc.  All rights reserved


import os
from xen.xend import sxp
import xen.xend.XendClient
from xen.xend.XendClient import server
from xen.xend.xenstore.xsutil import *


class Bridge:
    def __init__ (self, nm): 
        self.ifaces = []
        self.name = nm 

    def add_iface (self, iface):
        #print "Adding interface " + iface + " to bridge " + self.name
        self.ifaces.append (iface)

    def iface_present (self, iface):
        return iface in self.ifaces

    def empty(self):
        return len(self.ifaces) == 0


def get_bridges():
    p = os.popen ("brctl show", "r")
    bridges = []
    bridge = None

    for line in p:
        if line[0] != '\t':
            if line.startswith ("bridge name"):
                continue
            parts = line.split ()
            name = parts[0]
            bridge = Bridge (name)
        
            if len(parts) == 4:         # interface present
                bridge.add_iface (parts[3])
            bridges.append (bridge)
        else: 
            ifacename = line.strip()
            bridge.add_iface (ifacename)

    return bridges


# returns a 3-tuple: bridgename, newbridge, addvnic
def get_bridge_name(vnic):
    bridges = get_bridges()

    # check if the vnic is part of a bridge
    for bridge in bridges:
        if bridge.iface_present (vnic):
             return bridge.name, False, False

    # no vnic present in a bridge, find an empty one or make a new one

    for bridge in bridges:
        if bridge.empty():
            return bridge.name, False, True

    # no empty bridges, make a new one
    # they have to be called xenbr<n> for the Xen Source management to pick them
    # up
    prefix = "xenbr%d"
    num = 0
    while num < 100:
        brname = prefix % num
        found = False
        for b in bridges:
            if brname == b.name:
                found = True
                break
        if not found:
            return brname, True, True
        num = num + 1
    raise Exception("Unable to make bridge for vnic %s" % vnic)


def get_bridge_interfaces (bridge):
    bridges = get_bridges()

    for b in bridges:
        if b.name == bridge:
            return b.ifaces
    return []

def getDomains(domain_names):
    if domain_names:
        return map(server.xend.domain, domain_names)
    else:
        return server.xend.domains(1)

def get_domain_names(doms):
    names = []
    for dom in doms:
        names.append (sxp.child_value(dom, 'name', 'unknown'))
    return names


def find_domain (doms, nameorid):
    for dom in doms:
        name = sxp.child_value(dom, 'name', 'unknown')
        id = sxp.child_value(dom, 'domid', 'unknown')

        if nameorid == name or nameorid == str(id):
            return dom

    return None


def find_domain_name (doms, nameorid):
    for dom in doms:
        name = sxp.child_value(dom, 'name', 'unknown')
        id = sxp.child_value(dom, 'domid', 'unknown')

        if nameorid == name or nameorid == str(id):
            return name

    return nameorid


def get_vnic_mac (vnic):
    p = os.popen ("ifconfig '%s'" % vnic)
    for line in p:
        if "HWaddr" in line:
            i = line.find ("HWaddr") + 7
            return line[i:].strip()
    return ""

# traverse the domain to find the mac address of a device on the given bridge
def get_mac (dom, bridge):
    for elem in dom:
        if elem[0] == 'device':
            found = False
            mac = ""
            for dnode in elem[1]:
                if dnode[0] == 'bridge' and dnode[1] == bridge:
                    found = True
                if dnode[0] == 'mac':
                    mac = dnode[1]
            if found:
                return mac
    return ""
        

def is_vnic(name):
    for root, dirs, files in os.walk ("/proc/driver/vnic/devices"):
        if name in files:
            return True
    return False


def make_vnic_bridge (vnic):
    from xen.util import Brctl

    # get an existing bridge, or make a new one
    bridge, newbridge, addvnic = get_bridge_name(vnic)

    if newbridge:
        Brctl.bridge_create (bridge)
    if addvnic:
        Brctl.vif_bridge_add ({'bridge':bridge, 'vif': vnic})

    return bridge


def list_vnics():
    doms = getDomains("")
    print "vnic             status     domains"
    print "-----------------------------------"
    for root, dirs, files in os.walk ("/proc/driver/vnic/devices"):
        for vnic in files:
            f = open ("/proc/driver/vnic/devices/" + vnic, 'r')
            up = True 
            connected = True
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                (tag, value) = line.split(':')
                if "Tx/Rx handle" in tag:
                    if value.strip() != "Connected":
                        connected = False 
                if "Control handle" in tag:
                    if value.strip() != "Connected":
                        connected = False 
                if tag == "VNIC up":
                    if value.strip() == "No":
                        up = False

            bridge, newbridge, addvnic = get_bridge_name (vnic)
            ifaces = get_bridge_interfaces (bridge)

            status = ""
            if connected and up:
                status = "in-use"
            else:
                 if not connected:
                    status = "down"
                 else: 
                    status = "available"

            domain_names = ""
            comma = False
            # now that we have the interface names, work out what domains they belong to
            for iface in ifaces:
                if "vif" in iface:
                    dot = iface.find('.')
                    domid = iface[3:dot]
                    name = find_domain_name (doms, domid)
                    if comma:
                        domain_names += ','
                    domain_names += "%s(%s)" % (name, domid)
                    comma = True
            print "%-16s %-10s %s" % (vnic, status, domain_names)


def get_vhba_devices (vhba):
    devices = []
    p = os.popen ("sg_map -x -sd")
    for line in p:
        try:
            (sg, host, bus, scsi_id, lun, scsi_type, dev) = line.split()
            devices.append ([host, dev])
        except Exception,e:
            pass

    vdevs = []

    f = open ("/proc/driver/vhba/devices/" + vhba, 'r')
    for line in f:
        line = line.strip()
        if len(line) == 0:
            continue
        if ':' in line:
            (tag, value) = line.split(':')
            if tag.strip() == "Host Number":
                hostnum = value.strip()
                devnames = []
                for dev in devices:
                    if dev[0] == hostnum:
                        vdevs.append (dev[1])

    return vdevs



#
# this function lists the vhba information.  The device names are read from the xenstore database.  
#

def list_vhbas():
    doms = getDomains("")
    # find the devices that are handled by vhbas

    devices = []
    p = os.popen ("sg_map -x -sd")
    for line in p:
        try:
            (sg, host, bus, scsi_id, lun, scsi_type, dev) = line.split()
            devices.append ([host, dev])
        except Exception,e:
            pass

    print "VHBA supplied devices"
    print "====================="
    print ""
    print "vhba    devices"
    print "---------------"

    # two maps are held, one mapping a vhba to a vector of its devices and the other
    # mapping a device to the vhba that owns it
    vhbamap = {}
    devmap = {}

    for root, dirs, files in os.walk ("/proc/driver/vhba/devices"):
        for vhba in files:
            f = open ("/proc/driver/vhba/devices/" + vhba, 'r')
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                if ':' in line:
                    (tag, value) = line.split(':')
                    if tag.strip() == "Host Number":
                        hostnum = value.strip()
                        devnames = []
                        for dev in devices:
                            if dev[0] == hostnum:
                                devnames.append (dev[1])
                                devmap[dev[1]] = vhba
                        vhbamap[vhba] = devnames

    # print out all vhbas and their devices
    for vhba in vhbamap.items():
        devstring = ""
        for dev in vhba[1]:
            devstring += dev + " "
        print "%-7s %s" % (vhba[0], devstring)

    print ""

    # now that we know what devices are owned by the vhba module, lets work out what
    # guests are connected to the vhbas

    # this information is held in the xenstore
    xs = xshandle()
    rootpath = xs.get_domain_path (0)       # domain 0

    print "VHBAs in use"
    print "============"
    print ""
    print "vhba    domain            phy-device   guest-device"
    print "---------------------------------------------------"

    # OK, this deserves explanation
    # the xenstore keeps information about all the domains.  It is like a file system.  The information
    # we are after is held in the domain 0 section of the database (see http://wiki.xensource.com/xenwiki/XenStoreReference)
    # the 'dev' part of the 'vdb' tree contains the guests device name.  The 'params' part contains the domain 0 part.  
    # Here's an example
    #
    # we map /dev/sdd to /dev/hdd in guest 3.  The virtual device id is 5696 (obtained from xm block-list 3)
    #
    # >>> tx = xs.transaction_start()
    # >>> xs.read (tx, "/local/domain/0/backend/vbd/3/5696/dev")
    # '/dev/hdd'
    # >>> xs.read (tx, "/local/domain/0/backend/vbd/3/5696/params")
    # 'sdd'


    tx = xs.transaction_start()
    for dom in doms:
        id = sxp.child_value(dom, 'domid', 'unknown')
        name = sxp.child_value(dom, 'name', 'unknown')
        devs = server.xend.domain.getDeviceSxprs(id, 'vbd')
        for dev in devs:
            devid = dev[0]
            path = "%s/backend/vbd/%s/%s/dev" % (rootpath, id, devid)
            guestdevname = xs.read (tx, path)

            path = "%s/backend/vbd/%s/%s/params" % (rootpath, id, devid)
            dom0devname = xs.read (tx, path)

            if not dom0devname == None:
                if not dom0devname.startswith ("/dev"):
                    dom0devname = "/dev/" + dom0devname

                if dom0devname in devmap:
                    vhba = devmap[dom0devname]
                    print "%-7s %-17s %-12s %s" % (vhba, name, dom0devname, guestdevname)
                
    xs.transaction_end (tx)


def get_vhba_info (vhba):
    doms = getDomains("")
    # find the devices that are handled by vhbas

    devnames = get_vhba_devices (vhba)

    # now that we know what devices are owned by the vhba module, lets work out what
    # guests are connected to the vhbas

    # this information is held in the xenstore
    xs = xshandle()
    rootpath = xs.get_domain_path (0)       # domain 0

    # see list_vhbas() above for details of this code

    vdevnames = []          # names of the virtual devices

    tx = xs.transaction_start()
    for dom in doms:
        id = sxp.child_value(dom, 'domid', 'unknown')
        name = sxp.child_value(dom, 'name', 'unknown')
        devs = server.xend.domain.getDeviceSxprs(id, 'vbd')
        for dev in devs:
            devid = dev[0]
            path = "%s/backend/vbd/%s/%s/dev" % (rootpath, id, devid)
            guestdevname = xs.read (tx, path)

            path = "%s/backend/vbd/%s/%s/params" % (rootpath, id, devid)
            dom0devname = xs.read (tx, path)

            if not dom0devname == None:
                if not dom0devname.startswith ("/dev"):
                    dom0devname = "/dev/" + dom0devname

                if dom0devname in devnames:
                    vdevnames.append (guestdevname)
                
    xs.transaction_end (tx)

    return vdevnames
  
# convert x to base 26 
def base26 (x):
    result = ""
    while x > 25 :
        y = x / 26
        result = result + ("%c" % (y+0x60))
        x = x - (y * 26)
    result = result + ("%c" % (x + 0x61))
    return result

